//==============================================================================
//
//  OvenMediaEngine
//
//  Created by Hyunjun Jang
//  Copyright (c) 2018 AirenSoft. All rights reserved.
//
//==============================================================================
#include "item.h"

#include <base/ovlibrary/converter.h>
#include <unistd.h>

#include <regex>

#include "config_private.h"

#if defined(__APPLE__)
#	define HOST_NAME_MAX _POSIX_HOST_NAME_MAX
#endif

namespace cfg
{
	static ov::String MakeIndentString(int indent)
	{
		static std::map<int, ov::String> cache;

		auto item = cache.find(indent);

		if (item != cache.end())
		{
			return item->second;
		}

		ov::String indent_string;

		for (int count = 0; count < indent; count++)
		{
			indent_string += "    ";
		}

		cache[indent] = indent_string;

		return indent_string;
	}

	static const char *StringFromJsonValueType(Json::ValueType type)
	{
		switch (type)
		{
			case Json::ValueType::nullValue:
				return "null";

			case Json::ValueType::intValue:
				return "int";

			case Json::ValueType::uintValue:
				return "uint";

			case Json::ValueType::realValue:
				return "real";

			case Json::ValueType::stringValue:
				return "string";

			case Json::ValueType::booleanValue:
				return "boolean";

			case Json::ValueType::arrayValue:
				return "array";

			case Json::ValueType::objectValue:
				return "object";
		}

		return "unknown";
	}

	static const char *StringFromValueType(ValueType type)
	{
		switch (type)
		{
			case ValueType::Unknown:
				break;

			case ValueType::String:
				return "String";

			case ValueType::Integer:
				return "Integer";

			case ValueType::Boolean:
				return "Boolean";

			case ValueType::Float:
				return "Float";

			case ValueType::Text:
				return "Text";

			case ValueType::Attribute:
				return "Attribute";

			case ValueType::Element:
				return "Element";

			case ValueType::List:
				return "List";
		}

		return "Unknown";
	}

	Item::Item(const Item &item)
	{
		_tag_name = item._tag_name;

		for (auto &parse_item : item._parse_list)
		{
			_parse_list.emplace_back(parse_item.name, parse_item.is_parsed, parse_item.from_default, nullptr);
		}

		_parent = item._parent;
		_parsed = item._parsed;
	}

	Item::Item(Item &&item) noexcept
	{
		std::swap(_tag_name, item._tag_name);
		std::swap(_parse_list, item._parse_list);
		std::swap(_parent, item._parent);

		_parsed = item._parsed;
	}

	const ov::String Item::GetTagName() const
	{
		return _tag_name;
	}

	const Item *Item::GetParent(const ov::String &name) const
	{
		if (_parent != nullptr)
		{
			if (_parent->GetTagName() == name)
			{
				return _parent;
			}

			return _parent->GetParent(name);
		}

		return nullptr;
	}

	Item &Item::operator=(const Item &item)
	{
		_tag_name = item._tag_name;
		_parent = item._parent;
		_parse_list = item._parse_list;

		// To update pointers of ValueBase::_target
		MakeParseList();

		return *this;
	}

	void Item::Register(const ov::String &name, ValueBase *value, bool is_optional, bool is_conditional_optional, bool need_to_resolve_path, ValueBase::OptionalCallback conditional_optional_callback, ValueBase::ValidationCallback post_process_callback)
	{
		value->SetOptional(is_optional);
		if (is_conditional_optional)
		{
			OV_ASSERT2(conditional_optional_callback != nullptr);
			value->SetOptionalCallback(conditional_optional_callback);
		}
		else
		{
			OV_ASSERT2(conditional_optional_callback == nullptr);
		}
		value->SetValidationCallback(post_process_callback);
		value->SetNeedToResolvePath(need_to_resolve_path);

		bool is_parsed = false;
		bool from_default = true;

		for (auto old_value = _parse_list.begin(); old_value != _parse_list.end(); ++old_value)
		{
			if (old_value->name == name)
			{
				// restore some values before overwrite
				is_parsed = old_value->is_parsed;
				from_default = old_value->from_default;

				ValueBase *base = old_value->value.get();

				if (base != nullptr)
				{
					// The attributes of two values must be the same
					OV_ASSERT2(value->GetType() == base->GetType());
					OV_ASSERT2(value->IsOptional() == base->IsOptional());
					OV_ASSERT2(value->IsNeedToResolvePath() == base->IsNeedToResolvePath());
				}

				_parse_list.erase(old_value);
				break;
			}
		}

		_parse_list.emplace_back(name, is_parsed, from_default, value);
	}

	ValueBase *Item::FindValue(const ov::String &name, const ParseItem *parse_item_to_find)
	{
		for (auto &parse_item : _parse_list)
		{
			if (
				(parse_item.name == name) &&
				((parse_item.value != nullptr) && (parse_item.value->GetType() == parse_item_to_find->value->GetType())))
			{
				return parse_item.value.get();
			}
		}

		return nullptr;
	}

	bool Item::IsParsed(const void *target) const
	{
		MakeParseList();

		for (auto &parse_item : _parse_list)
		{
			OV_ASSERT2(parse_item.value != nullptr);

			if (parse_item.value->GetTarget() == target)
			{
				return parse_item.is_parsed;
			}
		}

		return false;
	}

	std::shared_ptr<ov::Error> Item::Parse(const ov::String &file_name, const ov::String &tag_name)
	{
		std::shared_ptr<ov::Error> error;

		if (ParseFromFile("", file_name, tag_name, 0, &error) != ParseResult::Error)
		{
			OV_ASSERT2(error == nullptr);
		}
		else
		{
			OV_ASSERT2(error != nullptr);
		}

		return error;
	}

	std::shared_ptr<ov::Error> Item::Parse(const ov::String &base_file_name, const Json::Value &json, const ov::String &tag_name)
	{
		std::shared_ptr<ov::Error> error;

		if (ParseFromJson(base_file_name, json, tag_name, 0, tag_name, &error) != ParseResult::Error)
		{
			OV_ASSERT2(error == nullptr);
		}
		else
		{
			OV_ASSERT2(error != nullptr);
		}

		return error;
	}

	bool Item::IsParsed() const
	{
		return _parsed;
	}

	bool Item::IsDefault(const void *target) const
	{
		MakeParseList();

		for (auto &parse_item : _parse_list)
		{
			OV_ASSERT2(parse_item.value != nullptr);

			if (parse_item.value->GetTarget() == target)
			{
				return parse_item.from_default;
			}
		}

		return false;
	}

	Item::ParseResult Item::ParseFromFile(const ov::String &base_file_name, ov::String file_name, const ov::String &tag_name, int indent, std::shared_ptr<ov::Error> *error)
	{
		_tag_name = tag_name;

		if ((ov::PathManager::IsAbsolute(file_name) == false) && (base_file_name.IsEmpty() == false))
		{
			file_name = ov::PathManager::Combine(ov::PathManager::ExtractPath(base_file_name), file_name);
		}

		pugi::xml_document document;
		pugi::xml_parse_result result = document.load_file(file_name);

		if (result == false)
		{
			*error = ov::Error::CreateError(
				"Config",
				"Could not read the file: %s (reason: %s)", file_name.CStr(), result.description());

			return ParseResult::Error;
		}

		return ParseFromNode(file_name, document.child(_tag_name), tag_name, indent, error);
	}

#define CONFIG_DECLARE_PROCESSOR(value_type, native_type, converter, process_value)                                                                                          \
	case value_type: {                                                                                                                                                       \
		auto target = dynamic_cast<Value<native_type> *>(parse_item.value.get());                                                                                            \
                                                                                                                                                                             \
		if ((target != nullptr) && (target->GetTarget() != nullptr))                                                                                                         \
		{                                                                                                                                                                    \
			auto temp_value = process_value;                                                                                                                                 \
			parse_item.is_parsed = true;                                                                                                                                     \
			if ((temp_value != nullptr) && (temp_value[0] != '\0'))                                                                                                          \
			{                                                                                                                                                                \
				*(target->GetTarget()) = converter(Preprocess(base_file_name, value, temp_value, name, indent));                                                             \
				logtd("%s<%s> <%s> = %s", MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr(), ToString(&parse_item, 0, true, false).CStr());                     \
				parse_item.from_default = false;                                                                                                                             \
			}                                                                                                                                                                \
			else                                                                                                                                                             \
			{                                                                                                                                                                \
				logtd("%s<%s> <%s> = %s (use default value)", MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr(), ToString(&parse_item, 0, true, false).CStr()); \
			}                                                                                                                                                                \
			result = ParseResult::Parsed;                                                                                                                                    \
		}                                                                                                                                                                    \
		else                                                                                                                                                                 \
		{                                                                                                                                                                    \
			OV_ASSERT2(false);                                                                                                                                               \
			return ParseResult::Error;                                                                                                                                       \
		}                                                                                                                                                                    \
                                                                                                                                                                             \
		break;                                                                                                                                                               \
	}

	// node는 this 레벨에 준하는 항목임. 즉, node.name() == _tag_name.CStr() 관계가 성립
	Item::ParseResult Item::ParseFromNode(const ov::String &base_file_name, const pugi::xml_node &node, const ov::String &tag_name, int indent, std::shared_ptr<ov::Error> *error)
	{
		_tag_name = tag_name;
		_parsed = false;

		MakeParseList();

		logtd("%s<%s> Trying to parse node (this = %p)...", MakeIndentString(indent).CStr(), _tag_name.CStr(), this);

		auto inc = node.attribute("include");

		if (inc.empty() == false)
		{
			logtd("%s<%s> Include file found: %s", MakeIndentString(indent).CStr(), _tag_name.CStr(), inc.value());

			std::vector<ov::String> file_list;
			auto path_error = ov::PathManager::GetFileList(base_file_name, inc.value(), &file_list);

			if (path_error != nullptr)
			{
				*error = ov::Error::CreateError(
					"Config",
					"<%s> Could not obtain file list: base_file: %s, include pattern: %s (%s)", _tag_name.CStr(), base_file_name.CStr(), inc.value(), path_error->ToString().CStr());
				return ParseResult::Error;
			}

			if (file_list.size() == 0)
			{
				*error = ov::Error::CreateError(
					"Config",
					"<%s> There is no file: base_file: %s, include pattern: %s", _tag_name.CStr(), base_file_name.CStr(), inc.value());
				return ParseResult::Error;
			}
			else if (file_list.size() > 1)
			{
				*error = ov::Error::CreateError(
					"Config",
					"<%s> Too many files found for single element: base_file: %s, include pattern: %s (%zu files found)", _tag_name.CStr(), base_file_name.CStr(), inc.value(), file_list.size());
				return ParseResult::Error;
			}

			auto result = ParseFromFile(base_file_name, inc.value(), tag_name, indent + 1, error);

			if (result == ParseResult::Error)
			{
				return ParseResult::Error;
			}

			_parsed = (result == ParseResult::Parsed);
		}

		// Check for invalid XML elements
		{
			for (auto &child : node.children())
			{
				auto name = child.name();

				auto item = std::find_if(_parse_list.begin(), _parse_list.end(), [&](const ParseItem &item) -> bool {
					return item.name == name;
				});

				if (item == _parse_list.end())
				{
					*error = ov::Error::CreateError(
						"Config",
						"Unknown configuration found: %s", GetSelector(child).CStr());
					return ParseResult::Error;
				}
			}
		}

		// Parse the items from node
		for (auto &parse_item : _parse_list)
		{
			auto &name = parse_item.name;
			ValueBase *value = parse_item.value.get();
			auto result = ParseResult::NotParsed;

			if (value == nullptr)
			{
				*error = ov::Error::CreateError(
					"Config",
					"<%s> Cannot obtain value for <%s> (this is a bug)", _tag_name.CStr(), name.CStr());
				return ParseResult::Error;
			}

			pugi::xml_node child_node;
			pugi::xml_attribute attribute;

			switch (value->GetType())
			{
				case ValueType::Text:
					child_node = node;
					break;

				case ValueType::Attribute:
					attribute = node.attribute(name);
					break;

				default:
					child_node = node.child(name);
					break;
			}

			if (
				((value->GetType() != ValueType::Attribute) && (child_node.empty() == false)) ||
				((value->GetType() == ValueType::Attribute) && (attribute.empty() == false)))
			{
				switch (value->GetType())
				{
					CONFIG_DECLARE_PROCESSOR(ValueType::Integer, int, ov::Converter::ToInt32, child_node.child_value())
					CONFIG_DECLARE_PROCESSOR(ValueType::Boolean, bool, ov::Converter::ToBool, child_node.child_value())
					CONFIG_DECLARE_PROCESSOR(ValueType::Float, float, ov::Converter::ToFloat, child_node.child_value())
					CONFIG_DECLARE_PROCESSOR(ValueType::String, ov::String, , child_node.child_value())
					CONFIG_DECLARE_PROCESSOR(ValueType::Text, ov::String, , child_node.child_value())
					CONFIG_DECLARE_PROCESSOR(ValueType::Attribute, ov::String, , attribute.value())

					case ValueType::Element: {
						auto target = dynamic_cast<ValueForElementBase *>(value);

						if ((target != nullptr) && (target->GetTarget() != nullptr))
						{
							target->GetTarget()->_parent = this;

							result = (target->GetTarget())->ParseFromNode(base_file_name, child_node, name, indent + 1, error);

							if (result == ParseResult::Error)
							{
								return result;
							}

							parse_item.is_parsed = (result == ParseResult::Parsed);
						}
						else
						{
							OV_ASSERT2(false);
							return ParseResult::Error;
						}

						logtd("%s<%s> <%s> = %s", MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr(), "{ ... }");

						break;
					}

					case ValueType::List: {
						logtd("%s<%s> Trying to parse item list <%s>...", MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr());

						auto target = dynamic_cast<ValueForListBase *>(value);

						if ((target != nullptr) && (target->GetTarget() != nullptr))
						{
							if ((child_node.empty() == false) && (_parsed == false))
							{
								// Clear only when _parsed is false, as there may be a list of included items.
								target->Clear();
							}

							while (child_node)
							{
								// Check the child has a include attribute
								auto inc = child_node.attribute("include");

								if (inc.empty() == false)
								{
									// Create items from the include files
									logtd("%s<%s> Include file pattern found: %s", MakeIndentString(indent).CStr(), _tag_name.CStr(), inc.value());

									std::vector<ov::String> file_list;
									auto path_error = ov::PathManager::GetFileList(base_file_name, inc.value(), &file_list);

									if (path_error != nullptr)
									{
										*error = ov::Error::CreateError(
											"Config",
											"<%s> Could not obtain file list: base_file: %s, include pattern: %s",
											_tag_name.CStr(),
											base_file_name.CStr(), inc.value());

										return ParseResult::Error;
									}

									// Sort by name
									std::sort(file_list.begin(), file_list.end());

									auto base_path = ov::PathManager::ExtractPath(base_file_name);

									for (auto file : file_list)
									{
										auto i = target->Create();

										i->_parent = this;

										logtd("%s<%s> Trying to parse node from file: %s", MakeIndentString(indent).CStr(), _tag_name.CStr(), file.CStr());

										// strip "base_path" from "file"
										auto file_name = file.Substring(base_path.GetLength());

										result = i->ParseFromFile(base_file_name, file, name, indent + 1, error);

										if (result == ParseResult::Error)
										{
											return result;
										}
									}

									parse_item.is_parsed = (result == ParseResult::Parsed);
								}
								else
								{
									auto i = target->Create();

									i->_parent = this;

									result = i->ParseFromNode(base_file_name, child_node, name, indent + 1, error);

									if (result == ParseResult::Error)
									{
										return result;
									}

									parse_item.is_parsed = (result == ParseResult::Parsed);
								}

								logtd("%s<%s> [List<%s>] = %s",
									  MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr(),
									  ov::String::FormatString("{ %zu items }", target->GetList().size()).CStr());

								child_node = child_node.next_sibling(name);
							}
						}
						else
						{
							OV_ASSERT2(false);
							return ParseResult::Error;
						}

						break;
					}

					case ValueType::Unknown:
						OV_ASSERT2(false);
						return ParseResult::Error;
				}
			}

			if (result == ParseResult::Parsed)
			{
				if (value->DoValidation() == false)
				{
					*error = ov::Error::CreateError(
						"Config",
						"Failed to validate %s.<%s>", GetSelector(node).CStr(), name.CStr());
					return ParseResult::Error;
				}

				parse_item.from_default = false;
			}

			// Check parsed from include file
			if (_parsed == false)
			{
				// It may parsed from XML, not include file
				if ((value->IsOptional() == false) && (result != ParseResult::Parsed))
				{
					logtd("%s<%s> Could not parse node <%s>", MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr());

					*error = ov::Error::CreateError(
						"Config",
						"%s.<%s> is required. Please check your configuration.", GetSelector(node).CStr(), name.CStr());
					return ParseResult::Error;
				}
			}
		}

		_parsed = true;
		return ParseResult::Parsed;
	}

#define CONFIG_DECLARE_PROCESSOR_FOR_JSON_PREPROCESS(value_type, native_type, type_check, process_value)                              \
	case value_type: {                                                                                                                \
		auto target = dynamic_cast<Value<native_type> *>(parse_item.value.get());                                                     \
                                                                                                                                      \
		if ((target != nullptr) && (target->GetTarget() != nullptr))                                                                  \
		{                                                                                                                             \
			if (process_value.isNull() == false)                                                                                      \
			{                                                                                                                         \
				if (type_check)                                                                                                       \
				{                                                                                                                     \
					parse_item.is_parsed = true;                                                                                      \
                                                                                                                                      \
					*(target->GetTarget()) = Preprocess(base_file_name, value, ov::Converter::ToString(process_value), name, indent); \
					logtd("%s%s.%s",                                                                                                  \
						  MakeIndentString(indent).CStr(),                                                                            \
						  path.CStr(), ToString(&parse_item, 0, true, false).CStr());                                                 \
					parse_item.from_default = false;                                                                                  \
				}                                                                                                                     \
				else                                                                                                                  \
				{                                                                                                                     \
					*error = ov::Error::CreateError(                                                                                  \
						"Config",                                                                                                     \
						"Invalid type for %s.%s: %s, expected: %s",                                                                   \
						path.CStr(), name.CStr(),                                                                                     \
						StringFromJsonValueType(process_value.type()), StringFromValueType(value_type));                              \
					return ParseResult::Error;                                                                                        \
				}                                                                                                                     \
			}                                                                                                                         \
			else                                                                                                                      \
			{                                                                                                                         \
				logtd("%s%s.%s (use default value)",                                                                                  \
					  MakeIndentString(indent).CStr(),                                                                                \
					  path.CStr(), ToString(&parse_item, 0, true, false).CStr());                                                     \
			}                                                                                                                         \
                                                                                                                                      \
			result = ParseResult::Parsed;                                                                                             \
		}                                                                                                                             \
		else                                                                                                                          \
		{                                                                                                                             \
			*error = ov::Error::CreateError("Config", "Internal server error");                                                       \
			OV_ASSERT2(false);                                                                                                        \
			return ParseResult::Error;                                                                                                \
		}                                                                                                                             \
                                                                                                                                      \
		break;                                                                                                                        \
	}

#define CONFIG_DECLARE_PROCESSOR_FOR_JSON(value_type, native_type, type_check, converter, process_value) \
	case value_type: {                                                                                   \
		auto target = dynamic_cast<Value<native_type> *>(parse_item.value.get());                        \
                                                                                                         \
		if ((target != nullptr) && (target->GetTarget() != nullptr))                                     \
		{                                                                                                \
			if (process_value.isNull() == false)                                                         \
			{                                                                                            \
				if (type_check)                                                                          \
				{                                                                                        \
					parse_item.is_parsed = true;                                                         \
                                                                                                         \
					*(target->GetTarget()) = converter(process_value);                                   \
					logtd("%s%s.%s",                                                                     \
						  MakeIndentString(indent).CStr(),                                               \
						  path.CStr(), ToString(&parse_item, 0, true, false).CStr());                    \
					parse_item.from_default = false;                                                     \
				}                                                                                        \
				else                                                                                     \
				{                                                                                        \
					*error = ov::Error::CreateError(                                                     \
						"Config",                                                                        \
						"Invalid type for %s.%s: %s, expected: %s",                                      \
						path.CStr(), name.CStr(),                                                        \
						StringFromJsonValueType(process_value.type()), StringFromValueType(value_type)); \
					return ParseResult::Error;                                                           \
				}                                                                                        \
			}                                                                                            \
			else                                                                                         \
			{                                                                                            \
				logtd("%s%s.%s (use default value)",                                                     \
					  MakeIndentString(indent).CStr(),                                                   \
					  path.CStr(), ToString(&parse_item, 0, true, false).CStr());                        \
			}                                                                                            \
                                                                                                         \
			result = ParseResult::Parsed;                                                                \
		}                                                                                                \
		else                                                                                             \
		{                                                                                                \
			*error = ov::Error::CreateError("Config", "Internal server error");                          \
			OV_ASSERT2(false);                                                                           \
			return ParseResult::Error;                                                                   \
		}                                                                                                \
                                                                                                         \
		break;                                                                                           \
	}

	Item::ParseResult Item::ParseFromJson(const ov::String &base_file_name, const Json::Value &json, const ov::String &tag_name, int indent, ov::String path, std::shared_ptr<ov::Error> *error)
	{
		_tag_name = tag_name;
		_parsed = false;

		MakeParseList();

		logtd("%s\"%s\" Trying to parse JSON (this = %p)...", MakeIndentString(indent).CStr(), _tag_name.CStr(), this);

		// Check for invalid JSON values
		{
			if (json.isObject())
			{
				for (auto child = json.begin(); child != json.end(); ++child)
				{
					auto name = child.name();

					if ((name.empty() == false) && (name[0] == '$'))
					{
						// This value is an attribute
						continue;
					}

					auto item = std::find_if(_parse_list.begin(), _parse_list.end(), [&](const ParseItem &item) -> bool {
						return item.name == name.c_str();
					});

					if (item == _parse_list.end())
					{
						*error = ov::Error::CreateError(
							"Config",
							"Unknown configuration found: %s.%s", path.CStr(), name.c_str());
						return ParseResult::Error;
					}
				}
			}
		}

		// Parse the items from JSON value
		for (auto &parse_item : _parse_list)
		{
			auto name = parse_item.name;
			ValueBase *value = parse_item.value.get();
			auto result = ParseResult::NotParsed;

			if (value == nullptr)
			{
				// If this happened, it's a bug
				*error = ov::Error::CreateError(
					"Config",
					"\"%s\" Cannot obtain value for \"%s\"", _tag_name.CStr(), name.CStr());
				OV_ASSERT2(false);
				return ParseResult::Error;
			}

			Json::Value child_json;

			switch (value->GetType())
			{
				case ValueType::Text:
					child_json = json;
					break;

				case ValueType::Attribute:
					name.Prepend("$");

					if (json.isObject())
					{
						child_json = json.get(name, Json::nullValue);
					}
					else
					{
						if (value->IsOptional())
						{
							child_json = Json::nullValue;
						}
						else
						{
							*error = ov::Error::CreateError(
								"Config",
								"Invalid type for %s.%s: %s, expected: %s",
								path.CStr(), name.CStr(),
								StringFromJsonValueType(json.type()), StringFromValueType(value->GetType()));
							return ParseResult::Error;
						}
					}
					break;

				case ValueType::List:
					if (json.isArray())
					{
						// Child JSONs are directly referenced in array form
						//
						// For example,
						//
						// XML:
						// <IceCandidates>
						//    <IceCandidate>*:10006-10010/udp</IceCandidate>
						// </IceCandidates>
						// => Both of IceCandidates/IceCandidate are ValueType::Element
						//
						// JSON:
						// {
						//     "IceCandidates": [ "*:10006-10010/udp" ]
						// }
						child_json = json;
					}
					else if (json.isObject())
					{
						child_json = json.get(name.CStr(), Json::nullValue);
					}
					else
					{
						*error = ov::Error::CreateError(
							"Config",
							"Invalid type for %s.%s: %s, expected: %s",
							path.CStr(), name.CStr(),
							StringFromJsonValueType(json.type()), StringFromValueType(value->GetType()));
						return ParseResult::Error;
					}
					break;

				default:
					
					if (json.isObject() == false)
					{
						*error = ov::Error::CreateError(
							"Config",
							"Invalid type for %s.%s: %s, expected: %s",
							path.CStr(), name.CStr(),
							StringFromJsonValueType(json.type()), StringFromValueType(value->GetType()));
						return ParseResult::Error;
					}
					child_json = json.get(name, Json::nullValue);
					break;
			}

			if (child_json.isNull() == false)
			{
				switch (value->GetType())
				{
					CONFIG_DECLARE_PROCESSOR_FOR_JSON(ValueType::Integer, int, child_json.isIntegral(), ov::Converter::ToInt32, child_json)
					CONFIG_DECLARE_PROCESSOR_FOR_JSON(ValueType::Boolean, bool, child_json.isBool(), ov::Converter::ToBool, child_json)
					CONFIG_DECLARE_PROCESSOR_FOR_JSON(ValueType::Float, float, child_json.isDouble(), ov::Converter::ToFloat, child_json)
					CONFIG_DECLARE_PROCESSOR_FOR_JSON_PREPROCESS(ValueType::String, ov::String, child_json.isString(), child_json)
					CONFIG_DECLARE_PROCESSOR_FOR_JSON_PREPROCESS(ValueType::Text, ov::String, child_json.isString(), child_json)
					CONFIG_DECLARE_PROCESSOR_FOR_JSON_PREPROCESS(ValueType::Attribute, ov::String, child_json.isString(), child_json)

					case ValueType::Element: {
						auto target = dynamic_cast<ValueForElementBase *>(value);
						auto item_target = target->GetTarget();

						if ((target != nullptr) && (item_target != nullptr))
						{
							item_target->_parent = this;

							if (child_json.isObject() == false)
							{
								// Check if the child_json is a processable type
								if (child_json.isString())
								{
									// Give a change to parse the child_json as ValueType::Text
									//
									// For example,
									//
									// XML:
									// <RTMP>
									//    <Port>80</Port>
									// </RTMP>
									// => Port is ValueType::Elemet
									//
									// JSON:
									// { "RTMP": {"Port": "80"} }
									// => This is the same configuration, but Port is a string
								}
								else if (child_json.isArray())
								{
									// Give a change to parse the child_json as ValueType::List
									//
									// For example,
									//
									// XML:
									// <IceCandidates>
									//    <IceCandidate>*:10006-10010/udp</IceCandidate>
									// </IceCandidates>
									// => Both of IceCandidates/IceCandidate are ValueType::Element
									//
									// JSON:
									// {
									//     "IceCandidates": [ "*:10006-10010/udp" ]
									// }
									// => This is the same configuration, but IceCandiates is a list of string
								}
								else
								{
									*error = ov::Error::CreateError(
										"Config",
										"Invalid type for %s.%s: %s, expected: %s",
										path.CStr(), name.CStr(),
										StringFromJsonValueType(child_json.type()), StringFromJsonValueType(Json::ValueType::objectValue));

									return ParseResult::Error;
								}
							}

							result = item_target->ParseFromJson(base_file_name, child_json, name, indent + 1, path + "." + name, error);

							if (result == ParseResult::Error)
							{
								return result;
							}

							parse_item.is_parsed = (result == ParseResult::Parsed);
						}
						else
						{
							OV_ASSERT2(false);
							return ParseResult::Error;
						}

						logtd("%s\"%s\" \"%s\" = %s", MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr(), "{ ... }");

						break;
					}

					case ValueType::List: {
						logtd("%s\"%s\" Trying to parse item list \"%s\"...", MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr());

						auto target = dynamic_cast<ValueForListBase *>(value);

						if ((target != nullptr) && (target->GetTarget() != nullptr))
						{
							// clear the list
							target->Clear();

							if (child_json.isArray() == false)
							{
								*error = ov::Error::CreateError(
									"Config",
									"Invalid type for %s.%s: %s, expected: %s",
									path.CStr(), name.CStr(),
									StringFromJsonValueType(child_json.type()), StringFromJsonValueType(Json::ValueType::arrayValue));

								return ParseResult::Error;
							}
							else
							{
								for (auto iterator = child_json.begin(); iterator != child_json.end(); ++iterator)
								{
									// child_item = value of iterator
									auto &child_item = *iterator;

									auto i = target->Create();

									i->_parent = this;

									result = i->ParseFromJson(base_file_name, child_item, name, indent + 1, path + "." + name, error);

									if (result == ParseResult::Error)
									{
										return result;
									}

									parse_item.is_parsed = (result == ParseResult::Parsed);

									logtd("%s<%s> [List<%s>] = %s",
										  MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr(),
										  ov::String::FormatString("{ %zu items }", target->GetList().size()).CStr());
								}
							}
						}
						else
						{
							OV_ASSERT2(false);
							return ParseResult::Error;
						}

						break;
					}

					case ValueType::Unknown:
						OV_ASSERT2(false);
						return ParseResult::Error;
				}
			}

			if (result == ParseResult::Parsed)
			{
				if (value->DoValidation() == false)
				{
					*error = ov::Error::CreateError(
						"Config",
						"Failed to validate %s.%s", path.CStr(), name.CStr());
					return ParseResult::Error;
				}

				parse_item.from_default = false;
			}

			// Check parsed from include file
			if (_parsed == false)
			{
				// It may parsed from XML, not include file
				if ((value->IsOptional() == false) && (result != ParseResult::Parsed))
				{
					logtd("%s\"%s\" Could not parse JSON \"%s\"", MakeIndentString(indent).CStr(), _tag_name.CStr(), name.CStr());

					*error = ov::Error::CreateError(
						"Config",
						"%s.%s is required. Please check your configuration.", path.CStr(), name.CStr());
					return ParseResult::Error;
				}
			}
		}

		_parsed = true;
		return ParseResult::Parsed;
	}

	ov::String Item::GetEnv(const char *key, const char *default_value, bool *is_default_value)
	{
		auto env = std::getenv(key);

		if (env != nullptr)
		{
			return env;
		}

		if (::strcmp(key, "HOSTNAME") == 0)
		{
			ov::String hostname;

			if (hostname.SetCapacity(HOST_NAME_MAX))
			{
				if (::gethostname(hostname.GetBuffer(), hostname.GetCapacity()) == 0)
				{
					return std::move(hostname);
				}
			}
		}

		if ((default_value != nullptr) && (default_value[0] != '\0'))
		{
			// Use the default value
			if (is_default_value != nullptr)
			{
				*is_default_value = true;
			}

			return default_value;
		}

		// There is no default value
		return "";
	}

	ov::String Item::Preprocess(const ov::String &xml_path, const ValueBase *value_base, const char *value, const ov::String &tag_name, int indent)
	{
		// Preprocess for ${env:XXX} macro
		std::string str = value;
		std::string result;
		std::regex r(R"(\$\{env:([^}]*)\})");

		auto start = str.cbegin();
		std::sregex_iterator iterator = std::sregex_iterator(start, str.cend(), r);

		while (iterator != std::sregex_iterator())
		{
			std::smatch matches = *iterator;
			auto match = matches[0];
			auto token = matches[1];

			auto env_key = token.str();
			auto position = env_key.find_first_of(':');

			std::string default_value;
			bool is_default_value = false;

			if (position != std::string::npos)
			{
				default_value = env_key.substr(position + 1);
				env_key = env_key.substr(0, position);
			}

			auto key = env_key.c_str();
			ov::String value = GetEnv(key, default_value.c_str(), &is_default_value);

			// _tag_name: Tag name of current element
			// tag_name: Tag name of child element
			logtd("%s<%s> <%s> Configuration from env: %s = \"%s\"%s",
				  MakeIndentString(indent).CStr(), _tag_name.CStr(), tag_name.CStr(),
				  key, value.CStr(), is_default_value ? " (default value)" : "");

			result.append(start, match.first);
			result.append(value);

			start = match.second;

			++iterator;
		}

		result.append(start, str.cend());

		// Preprocess for ResolvePath annotation
		if (value_base->IsNeedToResolvePath())
		{
			const char *path = result.c_str();

			if (ov::PathManager::IsAbsolute(path) == false)
			{
				// relative path
				ov::String base_path = ov::PathManager::ExtractPath(xml_path);

				return ov::PathManager::Combine(base_path, path);
			}
		}

		return result.c_str();
	}

	ov::String Item::GetSelector(const pugi::xml_node &node)
	{
		std::vector<ov::String> selectors;

		selectors.emplace_back(node.name());

		auto parent = node.parent();

		while (
			(parent.type() != pugi::xml_node_type::node_null) &&
			(parent.type() != pugi::xml_node_type::node_document))
		{
			selectors.insert(selectors.begin(), parent.name());

			parent = parent.parent();
		}

		ov::String selector = "<";
		selector.Append(ov::String::Join(selectors, ">.<"));
		selector.Append('>');

		return selector;
	}

	ov::String Item::ToString(bool exclude_default) const
	{
		return ToString(0, exclude_default);
	}

	ov::String Item::ToString(int indent, bool exclude_default) const
	{
		ov::String result;

		MakeParseList();

		if (indent == 0)
		{
			result.AppendFormat("%s = ", _tag_name.CStr());
		}

		if (_parse_list.empty())
		{
			result.Append("{}");
		}
		else
		{
			result.Append("{\n");

			for (auto &parse_item : _parse_list)
			{
				result.Append(ToString(&parse_item, indent + 1, exclude_default, true));
			}

			result.AppendFormat("%s}", MakeIndentString(indent).CStr());
		}

		return result;
	}

	ov::String Item::ToString(const ParseItem *parse_item, int indent, bool exclude_default, bool append_new_line) const
	{
		ov::String str;

		ValueBase *value = parse_item->value.get();

		switch (value->GetType())
		{
			case ValueType::Integer: {
				auto target = dynamic_cast<Value<int> *>(value);
				str = (target != nullptr) ? ov::String::FormatString("%d", *(target->GetTarget())) : "(null)";
				break;
			}

			case ValueType::Boolean: {
				auto target = dynamic_cast<Value<bool> *>(value);
				str = (target != nullptr) ? (*(target->GetTarget()) ? "true" : "false") : "(null)";
				break;
			}

			case ValueType::Float: {
				auto target = dynamic_cast<Value<float> *>(value);
				str = (target != nullptr) ? ov::String::FormatString("%f", *(target->GetTarget())) : "(null)";
				break;
			}

			case ValueType::String:
			case ValueType::Attribute:
			case ValueType::Text: {
				auto target = dynamic_cast<Value<ov::String> *>(value);
				str = (target != nullptr) ? *(target->GetTarget()) : "(null)";
				break;
			}

			case ValueType::Element: {
				auto target = dynamic_cast<ValueForElementBase *>(value);
				str = (target != nullptr) ? target->GetTarget()->ToString(indent, exclude_default) : "(null)";
				break;
			}

			case ValueType::List: {
				auto target = dynamic_cast<ValueForListBase *>(value);

				ov::String indent_string = MakeIndentString(indent + 1);
				auto item_list = target->GetList();

				if (item_list.empty())
				{
					str = "(no items) []";
				}
				else
				{
					str = ov::String::FormatString("(%d items) [", item_list.size());
					bool is_first = true;

					for (auto &item : item_list)
					{
						if (item == nullptr)
						{
							OV_ASSERT2(false);
							return "";
						}

						if (is_first)
						{
							str.AppendFormat("%s", item->ToString(indent, exclude_default).CStr());
							is_first = false;
						}
						else
						{
							str.AppendFormat(", %s", item->ToString(indent, exclude_default).CStr());
						}
					}

					str.Append(']');
				}

				break;
			}

			default:
				break;
		}

		return ov::String::FormatString(
			"%s%s%s%s (%s%s%s%s%s) = %s%s",
			MakeIndentString(indent).CStr(),
			value->GetType() == ValueType::List ? "List<" : "",
			parse_item->name.CStr(),
			value->GetType() == ValueType::List ? ">" : "",
			StringFromValueType(value->GetType()),
			value->IsOptional() ? ", optional" : "",
			value->IsNeedToResolvePath() ? ", path" : "",
			parse_item->is_parsed ? ", parsed" : "",
			parse_item->from_default ? ", default" : "",
			exclude_default ? (parse_item->is_parsed ? str.CStr() : "N/A") : str.CStr(),
			append_new_line ? "\n" : "");
	}
};	// namespace cfg